/*
 * Copyright 2016 Steve Nickolas.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a 
 * copy of this software and associated documentation files (the "Software"), 
 * to deal with the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 *   1. Redistributions of source code must retain the above copyright notice, 
 *      this list of conditions and the following disclaimers.
 *   2. Redistributions in binary form must reproduce the above copyright 
 *      notice, this list of conditions and the following disclaimers in the 
 *      documentation and/or other materials provided with the distribution.
 *   3. Neither the name of the author, nor the names of its contributors, may 
 *      be used to endorse or promote products derived from this Software 
 *      without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS WITH THE SOFTWARE.
 */

#define DEFAULT_VOLUME 23

#include <stdint.h>
#include <stdio.h>
#include <windows.h>
#include <SDL/SDL.h>

__declspec(dllexport) uint32_t mod_type=0x524B5053;
__declspec(dllexport) uint8_t *mod_name="SDL Speaker Driver";
__declspec(dllexport) uint8_t *mod_pref="sndsdl";
__declspec(dllexport) uint8_t *mod_vers="0.01";
__declspec(dllexport) uint8_t *mod_copr="Copyright 2016 Steve Nickolas";

static uint8_t sndon, sndhalt;
static uint8_t polarity, amplitude;

SDL_AudioSpec wanted;

uint8_t ringbuffer[65536];
uint16_t ringposr, ringposw;

/*
 * Sound gets a bit weird at native rate of 15720,
 * so internally double it to 31440.
 */
void rebuffer (void *udata, Uint8 *stream, int len)
{
 int t;
 
 if (sndhalt)
 {
  memset(stream,0,len);
  return;
 }

 for (t=0; t<len; t+=2) 
 {
  /* underrun! */
  if (((int16_t)(ringposw-ringposr))<-500)
  {
   stream[t]=stream[t+1]=0;
  }
  else
  {
   stream[t]=stream[t+1]=ringbuffer[ringposr];
   /* burn behind us */
   ringbuffer[ringposr++]=0;
  }
 };
 
 /* if things get too laggy, skip ahead */
 if (((int16_t)(ringposw-ringposr))>4000)
 {
  while (ringposr!=ringposw-2000)
  {
   ringbuffer[ringposr++]=0;
  }
 }
}

__declspec(dllexport) int sndsdl_sndinit (void)
{
 if (SDL_Init(SDL_INIT_TIMER|SDL_INIT_AUDIO)<0)
 {
  fprintf (stderr, "SDL: Could not initialize SDL: %s\n", SDL_GetError());
  return -1;
 }
 
 //wanted.freq=31440;
 wanted.freq=28400;
 wanted.format=AUDIO_U8;
 wanted.channels=1;
 wanted.samples=1024;
 wanted.callback=rebuffer;
 wanted.userdata=0;
 
 if (SDL_OpenAudio(&wanted,0)<0)
 {
  fprintf (stderr, "SDL: Could not initialize sound driver: %s\n", SDL_GetError());
  return -1;
 }
 
 SDL_PauseAudio(0);
 
 memset(ringbuffer,0,65536);
 
 polarity=0;
 amplitude=DEFAULT_VOLUME;
 sndon=1;
 ringposw=0;
 ringposr=63500;
 return 0;
}

__declspec(dllexport) void sndsdl_snddeinit (void)
{
 SDL_CloseAudio();
}

__declspec(dllexport) void sndsdl_snddo (void)
{
 polarity=!polarity;
}

__declspec(dllexport) void sndsdl_sndenable (uint8_t enable)
{
 sndhalt=!enable;

 if (!enable)
 {
  memset(ringbuffer,0,65536);
  ringposw=0;
  ringposr=63500;
 }
}

//uint32_t debugiters;

__declspec(dllexport) void sndsdl_sndloop (void)
{
 if (!sndon) return;
 if (sndhalt) return;

 if (((int16_t)(ringposw-ringposr))<-500)
 {
  while (ringposr!=ringposw-500)
  {
   ringbuffer[--ringposw]=0;
  }
 }
 
 ringbuffer[ringposw++]=polarity*amplitude;
 
 /*
 debugiters++;
 if (debugiters==15720)
 {
  debugiters=0;
  printf ("Lag: %d samples\n", (int16_t)(ringposw-ringposr));
 }
 */
}

__declspec(dllexport) void sndsdl_sndconfig (void)
{
}
