#include <errno.h>
#include <stdio.h>
#include <string.h>

unsigned char bootsect[256]=
{
 0x01, 0xA5, 0x2B, 0x48, 0xAA, 0xBD, 0x88, 0xC0,
 0x20, 0x84, 0xFE, 0x20, 0x2F, 0xFB, 0x20, 0x93,
 0xFE, 0x20, 0x89, 0xFE, 0xAD, 0xB3, 0xFB, 0xC9,
 0x06, 0xF0, 0x08, 0xA9, 0x40, 0x85, 0x36, 0xA9,
 0x08, 0x85, 0x37, 0xA0, 0x00, 0xB9, 0x4D, 0x08,
 0xF0, 0x08, 0x09, 0x80, 0x20, 0xED, 0xFD, 0xC8,
 0xD0, 0xF3, 0x20, 0x89, 0xFE, 0x20, 0x0C, 0xFD,
 0x68, 0x4A, 0x4A, 0x4A, 0x4A, 0x4C, 0x95, 0xFE,
 0xC9, 0xE1, 0x90, 0x06, 0xC9, 0xFB, 0xB0, 0x02,
 0x29, 0xDF, 0x4C, 0xF0, 0xFD, 0x4E, 0x6F, 0x6E,
 0x2D, 0x53, 0x79, 0x73, 0x74, 0x65, 0x6D, 0x20,
 0x64, 0x69, 0x73, 0x6B, 0x20, 0x6F, 0x72, 0x20,
 0x64, 0x69, 0x73, 0x6B, 0x20, 0x65, 0x72, 0x72,
 0x6F, 0x72, 0x0D, 0x52, 0x65, 0x70, 0x6C, 0x61,
 0x63, 0x65, 0x20, 0x61, 0x6E, 0x64, 0x20, 0x70,
 0x72, 0x65, 0x73, 0x73, 0x20, 0x61, 0x6E, 0x79,
 0x20, 0x6B, 0x65, 0x79, 0x20, 0x77, 0x68, 0x65,
 0x6E, 0x20, 0x72, 0x65, 0x61, 0x64, 0x79, 0x0D,
 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};

unsigned char image[143360];

struct vtoc {
 unsigned char      reserved1;
 unsigned char      cattrack;           /* 0x11      */
 unsigned char      catsect;            /* 0x0F      */
 unsigned char      dosversion;         /* 0x03      */
 unsigned short int reserved2;
 unsigned char      volume;             /* 0xFE      */
 unsigned char      reserved3[32];
 unsigned char      tsentries;          /* 0x7A      */
 unsigned char      reserved4[8];
 unsigned char      lasttrack;          /* 0x11      */
 unsigned char      allocdir;           /* 0x01      */
 unsigned short int reserved5;
 unsigned char      tracks;             /* 0x23      */
 unsigned char      sectors;            /* 0x10      */
 unsigned char      bytespersector[2];  /* 0x00 0x01 */
 unsigned char      trkbmp[35][4];
} vtoc;

unsigned short int tralloc (int numsecs)
{
 unsigned short int tmp=0xFFFF;

 /* this only allocs contig. secs on a free track. */
 while (numsecs) {numsecs--; tmp<<=1;}

 return tmp;
}

int main (int argc, char **argv)
{
 FILE *file;
 int travel;
 long dossize;
 int dostrax;

 vtoc.cattrack=0x11;
 vtoc.catsect=0x0F;
 vtoc.dosversion=0x03;
 vtoc.volume=254;
 vtoc.tsentries=0x7A;
 vtoc.lasttrack=0x11;
 vtoc.allocdir=0x01;
 vtoc.tracks=35;
 vtoc.sectors=16;
 vtoc.bytespersector[0]=256&0xFF;
 vtoc.bytespersector[1]=256>>8;

 for (travel=0; travel<35; travel++)
 {
  vtoc.trkbmp[travel][0]=255;
  vtoc.trkbmp[travel][1]=255;
  vtoc.trkbmp[travel][2]=0;
  vtoc.trkbmp[travel][3]=0;
 }

 vtoc.trkbmp[0][0]=0;                   /* DOS can't use track 0 */
 vtoc.trkbmp[0][1]=0;
 vtoc.trkbmp[17][0]=0;
 vtoc.trkbmp[17][1]=0;

 memset(image, 0, 143360);
 memcpy(image, bootsect, 256);
 memcpy(&(image[(17<<12)]), &vtoc, sizeof vtoc);

 for (travel=2; travel<17; travel++)
 {
  image[(17<<12)+(travel<<8)+1]=0x11;
  image[(17<<12)+(travel<<8)+2]=travel-1;
 }

 if ((argc<2)||(argc>3))
 {
  fprintf (stderr, "usage: %s filename.dsk [filename.dos]\n", argv[0]);
  return -1;
 }

 if (argc==3)
 {
  file=fopen(argv[2], "rb");
  if (!file)
  {
   fprintf (stderr, "Could not read DOS image '%s' (%s);\n"
                    "  formatting without a DOS image.\n", argv[2], strerror(errno));
  }
  else
  {
   fseek(file,0,SEEK_END);
   dossize=ftell(file);
   dostrax=dossize>>12;
   fseek(file,0,SEEK_SET);
   if (dostrax>17)
   {
    fprintf (stderr, "DOS image '%s' is too large;\n"
                     "  formatting without a DOS image.\n", argv[2], strerror(errno));
   }
   else
   {
    unsigned int dossecs;

    fread(image, 1, dossize, file);
    fclose(file);

    dossecs=(dossize>>8)+((dossize&0xFF)?1:0);

    for (travel=0; travel<35; travel++)
    {
     if (dossecs>15)
     {
      vtoc.trkbmp[travel][0]=0;
      vtoc.trkbmp[travel][1]=0;
      dossecs-=16;
     }
     else if (dossecs)
     {
      unsigned int hold;

      hold=tralloc(dossecs);

      vtoc.trkbmp[travel][0]=hold>>8;
      vtoc.trkbmp[travel][1]=hold&0xFF;

      dossecs=0;
     }
    }

    printf("%d sectors allocated for DOS image '%s'.\n", dossize>>8, argv[2]);
   }
  }
 }

 memcpy(&(image[(17<<12)]), &vtoc, sizeof vtoc);

 file=fopen(argv[1], "wb");
 if (!file)
 {
  fprintf (stderr, "Could not open disk image '%s' (%s).\n", argv[1], strerror(errno));
  return 1;
 }

 fwrite(image, 1, 143360, file);
 fclose(file);

 return 0;
}
